#!/usr/bin/env python

# THIS IS DEBUG DEMONSTATION CODE ONLY

# https://github.com/victronenergy/venus/wiki/dbus#grid-and-genset-meter
# https://shelly-api-docs.shelly.cloud/gen2/0.14/ComponentsAndServices/Shelly
 
# import normal packages
import platform 
import logging
import sys
import os
import sys
if sys.version_info.major == 2:
    import gobject
else:
    from gi.repository import GLib as gobject
import sys
import time
import requests # for http GET
import configparser # for config/ini file
from requests.auth import HTTPDigestAuth
 
# our own packages from victron
sys.path.insert(1, os.path.join(os.path.dirname(__file__), '/opt/victronenergy/dbus-systemcalc-py/ext/velib_python'))
from vedbus import VeDbusService


class DbusShelly1pmService:
  def __init__(self, servicename, paths, productname='Shelly(Plus) 1PM', connection='Shelly(Plus) 1PM HTTP JSON service'):
    config = self._getConfig()
    deviceinstance = int(config['DEFAULT']['Deviceinstance'])
    customname = config['DEFAULT']['CustomName']
    plusPmSupport = config['DEFAULT']['PlusPmSupport']

    self._dbusservice = VeDbusService("{}.http_{:02d}".format(servicename, deviceinstance))
    self._paths = paths
    
    logging.debug("%s /DeviceInstance = %d" % (servicename, deviceinstance))
    
    # Create the management objects, as specified in the ccgx dbus-api document
    self._dbusservice.add_path('/Mgmt/ProcessName', __file__)
    self._dbusservice.add_path('/Mgmt/ProcessVersion', 'Unkown version, and running on Python ' + platform.python_version())
    self._dbusservice.add_path('/Mgmt/Connection', connection)
    
    # Create the mandatory objects
    self._dbusservice.add_path('/DeviceInstance', deviceinstance)
    self._dbusservice.add_path('/ProductId', 0xFFFF) 
    self._dbusservice.add_path('/ProductName', productname)
    self._dbusservice.add_path('/CustomName', customname)    
    self._dbusservice.add_path('/Connected', 1)
    
    self._dbusservice.add_path('/Latency', None)    
    self._dbusservice.add_path('/FirmwareVersion', 0.1)
    self._dbusservice.add_path('/HardwareVersion', 0)

    self._dbusservice.add_path('/Mode', 2)  # Switch position: 2=Inverter on; 4=Off; 5=Low Power/ECO
    self._dbusservice.add_path('/State', 9) # 0=Off; 1=Low Power; 2=Fault; 9=Inverting

    self._dbusservice.add_path('/Serial', self._getShellySerial())
    self._dbusservice.add_path('/UpdateIndex', 0)
    self._dbusservice.add_path('/StatusCode', 0)  # Dummy path so VRM detects us as a PV-inverter.
    
    # add path values to dbus
    for path, settings in self._paths.items():
      self._dbusservice.add_path(
        path, settings['initial'], gettextcallback=settings['textformat'], writeable=True, onchangecallback=self._handlechangedvalue)

    # last update
    self._lastUpdate = 0

    # add _update function 'timer'
    gobject.timeout_add(250, self._update) # pause 250ms before the next request
    
    # add _signOfLife 'timer' to get feedback in log every 5minutes
    gobject.timeout_add(self._getSignOfLifeInterval()*60*1000, self._signOfLife)
 
  def _getShellySerial(self):
    config = self._getConfig()                                                                                                          
    accessType = config['DEFAULT']['AccessType']                                                                                        
    plusPmSupport = config['DEFAULT']['PlusPmSupport']  
    meter_data = self._getShellyData()  
    
#    if not meter_data['mac'] and not meter_data['sys']['mac']:
    if not meter_data['sys']['mac']:
        raise ValueError("Response does not contain 'mac' attribute")
    
    if plusPmSupport == 'True':
        serial = meter_data['sys']['mac']
    else:
        serial = meter_data['mac']
    return serial
 
 
  def _getConfig(self):
    config = configparser.ConfigParser()
    config.read("%s/config.ini" % (os.path.dirname(os.path.realpath(__file__))))
    return config;
 
 
  def _getSignOfLifeInterval(self):
    config = self._getConfig()
    value = config['DEFAULT']['SignOfLifeLog']
    
    if not value: 
        value = 0
    
    return int(value)
  
  
  def _getShellyStatusUrl(self):
    config = self._getConfig()
    accessType = config['DEFAULT']['AccessType']
    plusPmSupport = config['DEFAULT']['PlusPmSupport']

    if accessType == 'OnPremise' and plusPmSupport == 'True':
        URL = "http://%s/rpc/Shelly.GetStatus" % (config['ONPREMISE']['Host'])
        URL = URL.replace(":@", "")
    elif accessType == 'OnPremise':
        URL = "http://%s:%s@%s/status" % (config['ONPREMISE']['Username'], config['ONPREMISE']['Password'], config['ONPREMISE']['Host'])
        URL = URL.replace(":@", "")
    else:
        raise ValueError("AccessType %s is not supported" % (config['DEFAULT']['AccessType']))
    
    return URL
    
 
  def _getShellyData(self):
    config = self._getConfig()                                                                                                                        
    plusPmSupport = config['DEFAULT']['PlusPmSupport']

    URL = self._getShellyStatusUrl()
    if plusPmSupport == 'True' and config['ONPREMISE']['Username'] != '' and config['ONPREMISE']['Password'] != '':
        meter_r = requests.get(url = URL, auth=HTTPDigestAuth(config['ONPREMISE']['Username'], config['ONPREMISE']['Password']))
    else:
        meter_r = requests.get(url = URL)
    
    # check for response
    if not meter_r:
        raise ConnectionError("No response from Shelly(Plus) 1PM - %s" % (URL))
    
    meter_data = meter_r.json()     
    
    # check for Json
    if not meter_data:
        raise ValueError("Converting response to JSON failed")
    
    
    return meter_data
 
 
  def _signOfLife(self):
    logging.info("--- Start: sign of life ---")
    logging.info("Last _update() call: %s" % (self._lastUpdate))
    config = self._getConfig()        
    str(config['DEFAULT']['Phase'])  
    pvinverter_phase = str(config['DEFAULT']['Phase'])
    #logging.info("Last /Ac/Out/%s/P: %s" % (pvinverter_phase, self._dbusservice['/Ac/Out/' + pvinverter_phase + '/P']))
    logging.info("Last /Ac/%s/Power: %s" % (pvinverter_phase, self._dbusservice['/Ac/' + pvinverter_phase + '/Power']))
    logging.info("--- End: sign of life ---")
    return True
 
  def _update(self):   
    try:
       #get data from Shelly 1pm
       meter_data = self._getShellyData()
       
       config = self._getConfig()
       str(config['DEFAULT']['Phase'])
    
       pvinverter_phase = str(config['DEFAULT']['Phase'])
       plusPmSupport = str(config['DEFAULT']['PlusPmSupport'])

       #send data to DBus
       for phase in ['L1', 'L2', 'L3']:
         #pre = '/Ac/Out/' + phase
         pre = '/Ac/' + phase

         if phase == pvinverter_phase and plusPmSupport == 'True':
             #power = meter_data['switch:0']['apower']
             total = meter_data['switch:0']['aenergy']['total']
             voltage = meter_data['switch:0']['voltage']
             current = meter_data['switch:0']['current']
             #current = power / voltage
             power = voltage * current

             #self._dbusservice[pre + '/V'] = voltage
             #self._dbusservice[pre + '/I'] = current
             #self._dbusservice[pre + '/P'] = power
             self._dbusservice[pre + '/Voltage'] = voltage
             self._dbusservice[pre + '/Current'] = current
             self._dbusservice[pre + '/Power'] = power

         elif phase == pvinverter_phase:

             power = meter_data['meters'][0]['power']
             total = meter_data['meters'][0]['total']
             voltage = 230
             current = power / voltage
           
             #self._dbusservice[pre + '/V'] = voltage
             #self._dbusservice[pre + '/I'] = current
             #self._dbusservice[pre + '/P'] = power
             self._dbusservice[pre + '/Voltage'] = voltage
             self._dbusservice[pre + '/Current'] = current
             self._dbusservice[pre + '/Power'] = power
       
         # logging
         if phase == pvinverter_phase:
            #logging.debug("voltage (%s/V): %s" % (pre, self._dbusservice[pre + '/V']))
            #logging.debug("current (%s/I): %s" % (pre, self._dbusservice[pre + '/I']))
            #logging.debug("power   (%s/P): %s" % (pre, self._dbusservice[pre + '/P']))
            logging.debug("voltage (%s/Voltage): %s" % (pre, self._dbusservice[pre + '/Voltage']))
            logging.debug("current (%s/Current): %s" % (pre, self._dbusservice[pre + '/Current']))
            logging.debug("power   (%s/Power): %s" % (pre, self._dbusservice[pre + '/Power']))
            logging.debug("---");
       
       # increment UpdateIndex - to show that new data is available
       index = self._dbusservice['/UpdateIndex'] + 1  # increment index
       if index > 255:   # maximum value of the index
         index = 0       # overflow from 255 to 0
       self._dbusservice['/UpdateIndex'] = index

       #update lastupdate vars
       self._lastUpdate = time.time()              
    except Exception as e:
       logging.critical('Error at %s', '_update', exc_info=e)
       
    # return true, otherwise add_timeout will be removed from GObject - see docs http://library.isr.ist.utl.pt/docs/pygtk2reference/gobject-functions.html#function-gobject--timeout-add
    return True
 
  def _handlechangedvalue(self, path, value):
    logging.debug("someone else updated %s to %s" % (path, value))
    return True # accept the change

def getLogLevel():
  config = configparser.ConfigParser()
  config.read("%s/config.ini" % (os.path.dirname(os.path.realpath(__file__))))
  logLevelString = config['DEFAULT']['LogLevel']
  
  if logLevelString:
    level = logging.getLevelName(logLevelString)
  else:
    level = logging.INFO
    
  return level


def main():
  #configure logging
  logging.basicConfig(      format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S',
                            level=getLogLevel(),
                            handlers=[
                                logging.FileHandler("%s/current.log" % (os.path.dirname(os.path.realpath(__file__)))),
                                logging.StreamHandler()
                            ])
  
 
  try:
      logging.info("Start");
  
      from dbus.mainloop.glib import DBusGMainLoop
      # Have a mainloop, so we can send/receive asynchronous calls to and from dbus
      DBusGMainLoop(set_as_default=True)
     
      #formatting 
      _kwh = lambda p, v: (str(round(v, 2)) + ' kWh')
      _a = lambda p, v: (str(round(v, 2)) + ' A')
      _w = lambda p, v: (str(round(v, 1)) + ' W')
      _v = lambda p, v: (str(round(v, 1)) + ' V')  
     
      #start our main-service
      pvac_output = DbusShelly1pmService(
        #servicename='com.victronenergy.inverter',
        servicename='com.victronenergy.grid',
        paths={
          #'/Ac/Out/L1/V': {'initial': 0, 'textformat': _v},
          #'/Ac/Out/L1/I': {'initial': 0, 'textformat': _a},
          #'/Ac/Out/L1/P': {'initial': 0, 'textformat': _w},
          '/Ac/L1/Voltage': {'initial': 0, 'textformat': _v},
          '/Ac/L1/Current': {'initial': 0, 'textformat': _a},
          '/Ac/L1/Power': {'initial': 0, 'textformat': _w},
        })
     
      logging.info('Connected to dbus, and switching over to gobject.MainLoop() (= event based)')
      mainloop = gobject.MainLoop()
      mainloop.run()            
  except Exception as e:
    logging.critical('Error at %s', 'main', exc_info=e)
if __name__ == "__main__":
  main()
